/**
 * Angular directive to convert JSON into HTML table. Inspired by Brian Park's 
 * MIT Licensed "angular-json-human.js" which turns JSON to HTML tables.
 * 
 *  Extended to do schema inference and produce more regular tables 
 *  by Eben Haber at Couchbase.
 *  
 *  How to use this: you can pass to this directive a JS object or JSON text. If
 *  the object is an array, the code produces an HTML table with a column for each
 *  member of each element of the array. If the array contains similar objects, then
 *  a nice regular table is created. When subobjects are found, they are turned into
 *  sub-tables in a cell of their parent table.
 *  
 *  For example, I use this as part of a tabbed panel that can show query results as either
 *  formatted JSON text, using CodeMirror, a Table, or a Tree. The "isSelected" function
 *  indicates which tab is selected. The controller has an object called "lastResult" which
 *  includes the JSON string for a query result (lastResult.result).  
 *  
 *  The code appears as follows:

      <ui-codemirror ng-show="isSelected(1)" ui-codemirror-opts="resultsEditorOptions" ng-model="lastResult.result"></ui-codemirror>
	  <div style="height:60vh" ng-show="isSelected(2)" class="shadow_box ajtd-scrollable" mn-json-table="lastResult.result"></div>
	  <div style="height:60vh" ng-show="isSelected(3)" class="shadow_box ajtd-scrollable" mn-json-tree="lastResult.result"></div>

 * The directives mn-json-table and mn-json-tree permit showing the data different ways.
 */
/* global _, angular */
'use strict';
var tableModule = angular.module('mnJsonTableOrig', ['mnJsonTableOrig.tmpls']);


//allow recursive compilation, tables within tables, turtles all the way down


tableModule.factory('RecursionHelper', 
    [
     '$compile',
     function ($compile) {
       var RecursionHelper = {
           compile: function (element) {
             var contents = element.contents().remove();
             var compiledContents;
             return function (scope, element) {
               if (!compiledContents) {
                 compiledContents = $compile(contents);
               }

               compiledContents(scope, function (clone) {
                 element.append(clone);
               });

               var json = scope.json;

               if (!_.isUndefined(json)) 
               {
                 // the table code requires an array to traverse,
                 // so if we have a plain object, put it in an array
                 if (_.isPlainObject(json))
                   scope.json = [json];

                 var schema = {};
                 scope.schema = inferSchema(schema,json);
                 scope.numCols = Object.keys(scope.schema).length;
                 scope.numRows = json.length;
               }
               scope.isUndefined = function(obj) {return _.isUndefined(obj);};
               scope.strLen = function(obj) {if (_.isString(obj)) return obj.length; else return 0;};
               scope.toggleExpand = toggleExpand;
               //console.log("  recursively compiling json, element size: " + scope.numRows + ", " + scope.numCols);
               scope.getKeys = function(obj) {return(Object.keys(obj));};
             };
           }
       };
       return RecursionHelper;
     }
     ]);


//convenience function to return a string type indicator for a value


var getTypeNameForVal = function(value) {
  if (_.isString(value))
    return('S');
  else if (_.isNumber(value))
    return('N');	
  else if (_.isBoolean(value))
    return('B');
  // treat arrays of strings specially
  else if (_.isArray(value) && value.length > 0 && _.isString(value[0]))
    return('R');
  else if (_.isArray(value))
    return('A');
  else if (_.isPlainObject(value) && _.keys(value).length > 0)
    return('O');
  else
    return('U');
};


//these methods take a JSON object or array of objects and infers a schema, with 
//names and types for each field. We will only infer top level attribute names, for
//nested objects we will output the inner objects as trees


var addObjectToSchema = function(object, schema) {
  //console.log("Adding object to schema: " + JSON.stringify(object));
  for (var key in object) {
    // keys that start with "$" cause problems in Angular
    var typeName = getTypeNameForVal(object[key]);

    // if we haven't seen the field before, add it to the schema
    if (!(key in schema)) {
      schema[key] = typeName;
    }

    // otherwise, make sure the types are the same. If not, change
    // the type to an array of all types we've seen
    else {
      var curType = schema[key];
      if (_.isArray(curType)) { // already has multiple types
        if (_.indexOf(curType,typeName) == -1) // not there already
        {
          curType.push(typeName); // add to array
        }
      }
      else if (typeName !== schema[key])	{ // type is different
        schema[key] = [];
        schema[key].push(curType);
        schema[key].push(typeName);
      }
    }			
  }

};

var inferSchema = function(schema,source) {

  if (_.isArray(source)) for (var i = 0; i < source.length; i++) {
    addObjectToSchema(source[i],schema);
  }
  else if (_.isPlainObject)
    addObjectToSchema(source,schema);

  return(schema);
};

//
// toggle expansion of long strings
//

var toggleExpand = function(ev) {
  var textElem = ev.target.parentNode.children[0];
  //console.log("ToggleExpand, class: " + textElem.className + ", src: " + ev.target.textContent);
  if (textElem.className.indexOf("ajtd-hideContent") > -1)
  {
    ev.target.textContent = "(less...)";
    textElem.className = textElem.className.replace("ajtd-hideContent","ajtd-showContent");
  }
  else if (textElem.className.indexOf("ajtd-showContent") > -1)
  {
    ev.target.textContent = "(more...)";
    textElem.className = textElem.className.replace("ajtd-showContent","ajtd-hideContent");
  }
};

//this directive makes sure that we have json data to work with, then puts it
//into the json-table-root template, which encloses, inside a table, the jsonTableHelper 
//directive applied to the data. <table json-table-helper=json></table>

//json-table-helper, below, does recursive magic!


tableModule.directive('mnJsonTableOrig', ['RecursionHelper', function (RecursionHelper) {
  return {
    restrict: 'A',
    scope: { json: '=mnJsonTableOrig' },
    templateUrl: 'template/mn-json-table-orig.tmpl',
    link: function (scope, element, attrs) {
      scope.$watch('json', function (json) {
        
//        console.log(JSON.stringify(json));
        if (typeof json === 'string') {
          try {
            json = JSON.parse(json);
          } catch (e) {
          }
        }

        // if the object has an array length one, peek inside
        while (Array.isArray(json) && json.length == 1 && Array.isArray(json[0])) {
          json = json[0];
        }

        // the table code requires an array to traverse,
        // so if we have a plain object, put it in an array
        if (!Array.isArray(json))
          json = [json];

        var schema = {};
        scope.json = json;
        //scope.jsonStr = JSON.stringify(json);
        scope.schema = inferSchema(schema,json);
        scope.numCols = Object.keys(scope.schema).length;
        scope.numRows = json.length;
        scope.isUndefined = function(obj) {return _.isUndefined(obj);};
        scope.strLen = function(obj) {if (_.isString(obj)) return obj.length; else return 0;};
        scope.getKeys = function(obj) {return(Object.keys(obj));};
        scope.toggleExpand = toggleExpand;

        //console.log("\n\ntabularizing new json, element size: " + scope.numRows + ", " + scope.numCols);
      });
    }
  };
}]);

tableModule.directive('mnJsonTableOrigRecursion', ['RecursionHelper', function (RecursionHelper) {
  return {
    restrict: 'A',
    scope: { json: '=mnJsonTableOrigRecursion' },
    templateUrl: 'template/mn-json-table-orig.tmpl',
    compile: function(tElem) {
      return RecursionHelper.compile(tElem)
    }
  };
}]);


//this module exists only to put the templates we use into the template cache


angular.module('mnJsonTableOrig.tmpls', []).run(
    [
     '$templateCache',
     function ($templateCache) {
       'use strict';
       $templateCache
       .put('template/mn-json-table-orig.tmpl', 
           //'<h3>{{"full: table size: " +  numRows + ", width: " + numCols}}</h3>'+
           '<table class="ajtd-root ajtd-object-value ajtd-table">'+
           '<thead><th class=ajtd-column-header ng-repeat="key in getKeys(schema) track by $index">{{key}}</th></thead>' +
           '<tbody>' +
           //'<tr>{{"table size: " +  numRows + ", width: " + numCols}}</tr>' +
           // json should be an array, so have a table row for each element
           '<tr ng-repeat="(key, val) in json track by $index">' +
           //'<td>{{"rownum: " + key}}</td>' +
           
           // for each schema column, output the current object's value for that column
           '<td class="ajtd-cell" ng-repeat="k in getKeys(schema) track by $index">' + 

//         '<h4>{{"v: " + v + ", undefined(val[k]): " + isUndefined(val[k])}}</h4>' +

           // if the value is undefined, nothing, if primitive, output value, otherwise json-tree
           '<div ng-if="isUndefined(val[k])"></div>'+

           // on-array/non-objects, output the plain value
           '<div ng-if="schema[k]!==\'A\' && schema[k]!==\'O\' && !isUndefined(val[k]) && (schema[k] !== \'S\' || strLen(val[k]) <= 64)" class=ajtd-value>{{val[k]}}</div>'+
           // ...unless it is a long string, it which case we add "more"/"less" toggle
           '<div ng-if="schema[k]===\'S\' && strLen(val[k]) > 64 && !isUndefined(val[k])" class=ajtd-value><div class="ajtd-hideContent">{{val[k]}}</div><a ng-click="toggleExpand($event)">(more...)</a></div>'+
           // put arrays and objects into a subtable
           '<div ng-if="(schema[k]===\'A\' || schema[k]===\'O\') && !isUndefined(val[k])" mn-json-table-orig-recursion=val[k]></div>'+
           '<div ng-if="schema[k]===\'R\' && !isUndefined(val[k])"></div>'+
           '</td></tr>' + // end ng-repeat
//           '</tr ng-show="numRows > 100"><td colspan="{{numCols}}">{{schema.length}} Table size {{json.length}} truncated at 100 rows.</td></tr>' + 
           '</tbody></table>'
       );
     }
     ]);
