/**
 * Angular directive to convert JSON into HTML tree. Inspired by Brian Park's 
 * MIT Licensed "angular-json-human.js" which turns JSON to HTML tables.
 * 
 *  Extended for trees by Eben Haber at Couchbase.
 *  
 *  This class takes a JS object or JSON string, and displays it as an HTML
 *  table. Generally, it expects an array of something. If it's an array of objects,
 *  then each row corresponds to one object, and the columns are the union of all 
 *  fields of the objects. If an object doesn't have a field, that cell is blank.
 *  
 *  
 *  , which object members indented. This is similar to pretty-printing 
 *  JSON, but is more compact (no braces or commas), and permits using colors
 *  to highlight field names vs. values, and one line from the next.
 *  
 *  For usage, see the header to mn-json-table.
 */
/* global _, angular */
'use strict';
angular.module('mnJsonTable', []).directive('mnJsonTable', function () {
  return {
    restrict: 'A',
    scope: { data: '=mnJsonTable' },
    template: '<div></div>',
    link: function (scope, element) {
      
      scope.toggleExpand = toggleExpand;
      
      scope.$watch('data', function (json) {
        if (typeof json === 'string') {
          try {
                scope.json_length = json.length;
            json = JSON.parse(json);
          } catch (e) {
          }
        }
        
        // start with an empty div, if we have data convert it to HTML
        var content = "<div>{}</div>";
        if (json)
          content = '<div class="ajtd-root ajtd-type-array">' + makeHTMLtable(json) + "</div>";

        // set our element to use this HTML        
        element.html(content);
      });
    }
  };
});


//
// recursion in Angular is really inefficient, so we will use a javascript 
// routine to convert the object to an HTML representation. It's not true to 
// the spririt of Angular, but it was taking 10 seconds or more to render 
// a table with tens of thousands of cells
// 

var makeHTMLtable = function(object) {
  var result = '';
  
  // we expect an array of objects, which we turn into an HTML table. the first step is 
  // to create the set of columns, by looking at the fields of every object. If the array
  // only has primitives, then we'll output a single column table listing them.  
  if (_.isArray(object)) {
    // find the columns
    var itemsKeysToObject = {};
    _.forEach(object, function (item, index) {
      if (_.isPlainObject(item)) _.forEach(item, function (value, key) {
        itemsKeysToObject[key] = true;
      });
    });

    // if we found no columns, then we have an array of primitves (or arrays) and should output them
    // as a set of table cells
    if (_.size(itemsKeysToObject) == 0)  {
      result += '<table class="ajtd-root ajtd-object-value ajtd-value"><tbody>';

      for (var i=0; i < object.length; i++) {
        result += '<td>' + makeHTMLtable(object[i]) + '</td>'
      } 
      result += '</tbody></table>';    

      return(result);
    }
    
    // otherwise, we have an array of objects. Make a table whose columns are the union of all
    // fields in all the objects

    result += '<table class="ajtd-root ajtd-object-value ajtd-table"><thead>';
    _.forEach(itemsKeysToObject, function(value,key) {
      result += '<th class=ajtd-column-header>' + key +'</th>';        
    });
    result += '</thead><tbody>';

    // for each object in the array, output all the possible column values
    _.forEach(object, function (item, index) {
      result += '<tr>'; // new row for each object 
      _.forEach(itemsKeysToObject, function(b,key) {
        var value = item ? item[key] : null;
        result += '<td class="ajtd-cell">'; // start the row
        
        // for objects and arrays, make a recursive call
        if (_.isArray(value) || _.isPlainObject(value))
          result += makeHTMLtable(value);
        
        // for long strings, output an expandable cell
        else if (_.isString(value) && value.length > 64)
          result += '<div class=ajtd-value><div class="ajtd-hideContent">' + value + '</div><a onClick="toggleExpand(this)">(more...)</a></div>';

        // for everything else, just output the values
        else if (!_.isUndefined(value))
          result += '<div class=ajtd-value>' + value + ' </div>';
        
        // except undefined values, in which case output a nbsp
        else 
          result += '&nbsp';
        
        result += '</td>'; // end the row
      })

      result += '</tr>';
    }); 
    
    // done with array, close the table
    result += '</tbody></table>';    
  }
 
  
  // for an object, create a table and iterate over the fields
  else if (_.isPlainObject(object)) {
    // header columns for each field
    result += '<table class="ajtd-root ajtd-object-value ajtd-table"><thead>';
    _.forEach(object, function(value,key) {
      result += '<th class=ajtd-column-header>' + key +'</th>';        
    });
    result += '</thead><tbody><tr>';

    // now output values for each field
    _.forEach(object, function(value,key) {
      // for arrays and objects, we need a recursive call
      if (_.isArray(value) || _.isPlainObject(value))
        result += '<td class="ajtd-cell"><div>' + makeHTMLtable(value) + '</div></td>';
        
      // otherwise, for primitives, output key/value pair
      else
        result += '<td class="ajtd-cell"><div>' + value + '</span></div></td>';
    });
    // finish the table
    result += '</tr></tbody></table>';    

  }
  
  // it's also possible we were passed a primitive value, in which case just put it in a div
  
  else
    result += '<div class=ajtd-value>' + object + '</div>'
  
  return(result);
};


//
//toggle expansion of long strings
//

var toggleExpand = function(ev) {
  var textElem = ev.parentNode.children[0];
  //console.log("ToggleExpand, class: " + textElem.className + ", src: " + ev.textContent);
  if (textElem.className.indexOf("ajtd-hideContent") > -1)
  {
    ev.textContent = "(less...)";
    textElem.className = textElem.className.replace("ajtd-hideContent","ajtd-showContent");
  }
  else if (textElem.className.indexOf("ajtd-showContent") > -1)
  {
    ev.textContent = "(more...)";
    textElem.className = textElem.className.replace("ajtd-showContent","ajtd-hideContent");
  }
};

